﻿/*
  KeeModifiedBy Plugin
  Copyright (C) 2023 Dominik Reichl <dominik.reichl@t-online.de>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.Reflection;
using System.Text;
using System.Windows.Forms;

using KeePass.Forms;
using KeePass.Plugins;
using KeePass.UI;

using KeePassLib;

namespace KeeModifiedBy
{
	public sealed class KeeModifiedByExt : Plugin
	{
		internal const string KmbName = "KeeModifiedBy";

		internal const string KmbkUser = KmbName + "_User";
		internal const string KmbkMachine = KmbName + "_Machine";

		private IPluginHost m_host = null;
		private readonly KmbColumnProvider m_cp = new KmbColumnProvider();

		public override bool Initialize(IPluginHost host)
		{
			Terminate();
			if(host == null) { Debug.Assert(false); return false; }
			m_host = host;

			PwEntry.EntryTouched += KeeModifiedByExt.OnEntryTouched;
			GlobalWindowManager.WindowAdded += KeeModifiedByExt.OnWindowAdded;
			GlobalWindowManager.WindowRemoved += KeeModifiedByExt.OnWindowRemoved;
			m_host.ColumnProviderPool.Add(m_cp);

			return true;
		}

		public override void Terminate()
		{
			if(m_host == null) return;

			PwEntry.EntryTouched -= KeeModifiedByExt.OnEntryTouched;
			GlobalWindowManager.WindowAdded -= KeeModifiedByExt.OnWindowAdded;
			GlobalWindowManager.WindowRemoved -= KeeModifiedByExt.OnWindowRemoved;
			m_host.ColumnProviderPool.Remove(m_cp);

			m_host = null;
		}

		private static void OnEntryTouched(object sender, ObjectTouchedEventArgs e)
		{
			if(e == null) { Debug.Assert(false); return; }

			try
			{
				if(!e.Modified) return;

				PwEntry pe = (e.Object as PwEntry);
				if(pe == null) { Debug.Assert(false); return; }

				string strUser = null;
				try { strUser = Environment.UserName; }
				catch(Exception) { Debug.Assert(false); }

				string strMachine = null;
				try { strMachine = Environment.MachineName; }
				catch(Exception) { Debug.Assert(false); }

				if(string.IsNullOrEmpty(strUser))
					pe.CustomData.Remove(KmbkUser);
				else pe.CustomData.Set(KmbkUser, strUser);

				if(string.IsNullOrEmpty(strMachine))
					pe.CustomData.Remove(KmbkMachine);
				else pe.CustomData.Set(KmbkMachine, strMachine);
			}
			catch(Exception) { Debug.Assert(false); }
		}

		private static void OnWindowAdded(object sender, GwmWindowEventArgs e)
		{
			if(e == null) { Debug.Assert(false); return; }

			PwEntryForm ef = (e.Form as PwEntryForm);
			if(ef != null) ef.Shown += KeeModifiedByExt.OnPwEntryFormShown;
		}

		private static void OnWindowRemoved(object sender, GwmWindowEventArgs e)
		{
			if(e == null) { Debug.Assert(false); return; }

			PwEntryForm ef = (e.Form as PwEntryForm);
			if(ef != null) ef.Shown -= KeeModifiedByExt.OnPwEntryFormShown;
		}

		private static void OnPwEntryFormShown(object sender, EventArgs e)
		{
			try { OnPwEntryFormShownEx(sender); }
			catch(Exception) { Debug.Assert(false); }
		}

		private static void OnPwEntryFormShownEx(object sender)
		{
			PwEntryForm f = (sender as PwEntryForm);
			if(f == null) { Debug.Assert(false); return; }
			if(f.MultipleValuesEntryContext != null) return;

			PwEntry pe = f.EntryRef;
			if(pe == null) { Debug.Assert(false); return; }

			string strUser = pe.CustomData.Get(KmbkUser);
			string strMachine = pe.CustomData.Get(KmbkMachine);

			Label lbl = GetControl<Label>(f, "m_lblModifiedData");
			if(lbl == null) { Debug.Assert(false); return; }

			ListView lv = GetControl<ListView>(f, "m_lvHistory");
			if(lv == null) { Debug.Assert(false); return; }

			using(RtlAwareResizeScope rs = new RtlAwareResizeScope(lbl))
			{
				Rectangle rectAuto = lbl.Bounds;
				int wNew = ((lbl.RightToLeft == RightToLeft.Yes) ?
					(rectAuto.Right - lv.Left) : (lv.Right - rectAuto.Left));

				lbl.AutoSize = false;
				lbl.Size = new Size(wNew, rectAuto.Height);
			}

			lbl.AutoEllipsis = true;

			StringBuilder sb = new StringBuilder();
			sb.Append(lbl.Text ?? "?");
			sb.Append(" (by ");
			sb.Append(string.IsNullOrEmpty(strUser) ? "?" : strUser);
			sb.Append(" @ ");
			sb.Append(string.IsNullOrEmpty(strMachine) ? "?" : strMachine);
			sb.Append(')');
			string strText = sb.ToString();

			lbl.Text = strText;

			ToolTip tt = GetField<ToolTip>(f, "m_ttRect");
			if(tt != null) tt.SetToolTip(lbl, strText);
			else { Debug.Assert(false); }
		}

		private static T GetControl<T>(Form f, string strName)
			where T : Control
		{
			if((f == null) || string.IsNullOrEmpty(strName)) return null;

			Control[] v = f.Controls.Find(strName, true);
			if((v == null) || (v.Length != 1)) return null;

			return (v[0] as T);
		}

		private static T GetField<T>(object o, string strName)
			where T : class
		{
			if((o == null) || string.IsNullOrEmpty(strName)) return null;

			FieldInfo fi = o.GetType().GetField(strName, BindingFlags.NonPublic |
				BindingFlags.Instance);

			return ((fi != null) ? (fi.GetValue(o) as T) : null);
		}
	}
}
